:- use_module(library(chr)).

:- use_module(library(data/pair)).

% mem_of(Xs,X) :- memberchk(X,Xs).
mem_of([X|_],X) :- !.
mem_of([_|Xs],X) :- mem_of(Xs,X).

:- chr_constraint mul/3, add/3, param/2, binop/4
                , node/1, delay/4, top/1, edge/3
                , alpha/2, done/0.
                  % (?node, ?node, -node)
                % , add(?node, ?node, -node) 
                % , param(+number,-node)
                % , binop(+,?node, ?node,-node)
                % , node(-node)
                % , delay(?,?,-,?)
                % , top(+node)
                % , edge(?,?,+)
                % , alpha(?,?)
                % .


op(add,X,1,Y,1,X+Y).
op(mul,X,Y,Y,X,X*Y).

node(N) ==> N=n(ID,_,_), gensym(n,ID), alpha(ID,0).
param(X,N) <=> node(N), N=n(_,X,_).

binop(Op,n(IX,X,_),n(IY,Y,_),n(_,Z,DZ)) <=> 
   op(Op,X,ZdX,Y,ZdY,Expr), 
   delay(X,Y,Z,Expr), 
   edge(DZ,ZdX,IX), 
   edge(DZ,ZdY,IY).

delay(X,Y,Z,Expr) <=> ground(X), ground(Y) | Z is Expr.

add(XD,YD,ZD) <=> binop(add,XD,YD,ZD).
mul(XD,YD,ZD) <=> binop(mul,XD,YD,ZD).

add1(X,Y,Z) :- node(Z), add(X,Y,Z).
mul1(X,Y,Z) :- node(Z), mul(X,Y,Z).

alpha(ID,_), top(n(ID,_,DZ)), node(n(ID,_,_)) <=> DZ=1.

edge(DZ,ZdX,IX), alpha(IX,DX) <=> ground(DZ*ZdX) | DX1 is DX+DZ*ZdX, alpha(IX,DX1).
done \ alpha(IX,DX), node(n(IX,_,Deriv)) <=> Deriv=DX.

vec(N,Vals,Vec) :- length(Vec,N), maplist(param,Vals,Vec).

test(Vecs, VecVals,Cost) :-
   length(Vecs,2),
   maplist(vec(3),VecVals,Vecs),
   maplist(prod,Vecs,Prods),
   sum(Prods,Cost).

sum(Nodes, Sum) :-
   param(0,Zero),
   foldl(add1, Nodes, Zero, Sum).

prod(Nodes, Prod) :-
   param(1,One),
   foldl(mul1, Nodes, One, Prod).

:- chr_constraint map/2, accum/2, collect/1, not_exist_map/0.
map(X,V) \ map(X,W) <=> V=W.
accum(Pairs,Result) <=> not_exist_map | Result=Pairs.
map(X,V), accum(Pairs,Result) <=> accum([X-V|Pairs],Result).
collect(Pairs) <=> accum([],Pairs).
map(_,_) \ not_exist_map <=> fail.

mmap(X,Y,M1,M2) :- rb_lookup(X,Y,M1) -> M2=M1; rb_insert_new(M1,X,Y,M2).
mmap2(X,Y,M1,M2) :- rb_insert_new(M1,X,Y,M2) -> true; rb_lookup(X,Y,M1), M1=M2.
